/* -*- mode: C++; tab-width: 4; c-basic-offset: 4; -*- */

/* AbiCollab- Code to enable the modification of remote documents.
 * Copyright (C) 2005 by Martin Sevior
 * Copyright (C) 2006 by Marc Maurer <uwog@uwog.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "pd_Document.h"
#include "px_CR_SpanChange.h"
#include "px_CR_FmtMarkChange.h"  
#include "px_CR_SpanChange.h"
#include "px_CR_FmtMark.h"        
#include "px_CR_Span.h"
#include "px_CR_Glob.h"           
#include "px_CR_StruxChange.h"
#include "px_CR_ObjectChange.h"   
#include "px_CR_Strux.h"
#include "px_CR_Object.h"
#include "xap_App.h"
#include "pd_Style.h"
#include "xap_Frame.h"
#include "fv_View.h"
#include "ut_sleep.h"
#include "pt_Types.h"

#include "AbiCollab_Import.h"
#include "AbiCollab.h"
#include "AbiCollab_Export.h"
#include <backends/xp/Buddy.h>

ABI_Collab_Import::ABI_Collab_Import(AbiCollab* pAbiCollab, PD_Document* doc):
	m_pDoc(doc),
	m_pAbiCollab(pAbiCollab)
{
}

ABI_Collab_Import::~ABI_Collab_Import()
{
}

bool ABI_Collab_Import::_isOverlapping(UT_sint32 pos1, UT_sint32 length1, UT_sint32 pos2, UT_sint32 length2)
{
	if (pos1 == pos2)
		return true;
	else if (pos1 < pos2)
		return (pos1 + length1 - 1 >= pos2);
	else
		return (pos2 + length2 - 1 >= pos1);
}

void ABI_Collab_Import::_calculateCollisionSeqence(UT_sint32 iIncomingRemoteRev, const UT_UTF8String& sIncomingDocUUID, UT_sint32& iStart, UT_sint32& iEnd)
{
	UT_DEBUGMSG(("ABI_Collab_Import::_calculateCollisionSeqence() - iIncomingRemoteRev: %d\n", iIncomingRemoteRev));

	// initialization
	iStart = -1;
	iEnd = -1;

	ABI_Collab_Export* pExport = m_pAbiCollab->getExport();
	UT_return_if_fail(pExport);

	const UT_GenericVector<ChangeAdjust *>* pExpAdjusts = pExport->getAdjusts();
	UT_return_if_fail(pExpAdjusts);

	// worst case: the whole outgoing changerecord stack is the collision sequence
	iStart = 0;
	iEnd = pExpAdjusts->getItemCount();

	// scan back to find the changerecord in our export list the remote has seen,
	// maybe we can narrow the collision sequence down
	UT_sint32 i = 0;
	for (i = pExpAdjusts->getItemCount()-1; i >= 0; i--)
	{
	    ChangeAdjust * pChange = pExpAdjusts->getNthItem(i);
		if (pChange)
		{
		    UT_DEBUGMSG(("Looking at exported changerecord - rev: %d, pos: %d, length: %d adjust %d, queue pos: %d\n", 
						pChange->m_iCRNumber, pChange->m_iDocPos, pChange->m_iLength, pChange->m_iAdjust, i));

		    if (iIncomingRemoteRev >= pChange->m_iCRNumber)
		    {
				UT_DEBUGMSG(("Found the changerecord the remote side has already seen (iIncomingRemoteRev = %d)\n", iIncomingRemoteRev));

				// a new changerecord can't collide with a changerecord the 
				// remote side had already seen, hence the +1
				iStart = i+1;
				break;
		    }
		}
		else
			UT_ASSERT_HARMLESS(UT_SHOULD_NOT_HAPPEN);
	}

	// now move upward, so we kill off the bottom where changes are from the same document as the incoming cr (you can not collide with your own cr's!)
	for (; iStart < UT_sint32(pExpAdjusts->getItemCount()); iStart++)
	{
		ChangeAdjust * pChange = pExpAdjusts->getNthItem(iStart);
		if (strcmp(pChange->m_pDocUUID, sIncomingDocUUID.utf8_str()) != 0)
		{
			// not the same document anymore, we can stop
			break;
		}
		else
			UT_DEBUGMSG(("Killing off matching change: %d\n", iStart));
	}
}

UT_sint32 ABI_Collab_Import::_getIncomingAdjustmentForState(const UT_GenericVector<ChangeAdjust *>* pExpAdjusts, UT_sint32 iStart, UT_sint32 iEnd, UT_sint32 iIncomingPos, UT_sint32 iIncomingLength, const UT_UTF8String& sIncomingUUID, std::deque<int>& incAdjs)
{
	UT_DEBUGMSG(("ABI_Collab_Import::_getIncomingAdjustmentForState()\n"));
	UT_return_val_if_fail(pExpAdjusts, 0);

	UT_sint32 iAdjust = 0;
	for (UT_sint32 j = iEnd-1; j>=iStart; j--)
	{
		ChangeAdjust* pPrev = pExpAdjusts->getNthItem(j);
		if (strcmp(sIncomingUUID.utf8_str(), pPrev->m_pDocUUID) == 0)
		{
			if (pPrev->m_iAdjust > 0) // TODO: is this check correct?
			{
				UT_DEBUGMSG(("Looking at possible adjustment with queue pos: %d, -adjust: %d\n", j, -pPrev->m_iAdjust));
				if (pPrev->m_iAdjust > 0 &&
					static_cast<UT_sint32>(pPrev->m_iOrigDocPos) < iIncomingPos+iAdjust &&
					_isOverlapping(pPrev->m_iOrigDocPos, pPrev->m_iLength, iIncomingPos+iAdjust, iIncomingLength)) // if the position was in the middle of an insert done previously, then we only need to take the insertion adjust partially into account
				{
					UT_DEBUGMSG(("ADJUST OVERLAP DETECTED with queue pos: %d, pPrev->m_iOrigDocPos: %d, pPrev->m_iLength: %d, iIncomingPos: %d, iAdjust: %d\n", j, pPrev->m_iOrigDocPos, pPrev->m_iLength, iIncomingPos, iAdjust));
					iAdjust -= (iIncomingPos+iAdjust - pPrev->m_iOrigDocPos);
					incAdjs.push_front(iIncomingPos+iAdjust - pPrev->m_iOrigDocPos);
				}
				else if (static_cast<UT_sint32>(pPrev->m_iOrigDocPos) < iIncomingPos+iAdjust)
				{
					UT_DEBUGMSG(("ADJUSTMENT influenced normally by queue pos: %d\n", j));
					iAdjust -= pPrev->m_iAdjust;
					incAdjs.push_front(pPrev->m_iAdjust);
				}
				else
				{
					UT_DEBUGMSG(("no ADJUSTMENT influence by queue pos: %d, pPrev->m_iProgDocPos: %d, pPrev->m_iOrigDocPos\n", j, pPrev->m_iOrigDocPos, pPrev->m_iOrigDocPos));
					incAdjs.push_front(0);
				}
			}
		}
	}
	return iAdjust;
}

/*!
 * Scan back through the CR's we've emitted since this remote CR was sent
 * and see if any overlap this one.
 * return true if there is a collision.
 */
bool ABI_Collab_Import::_checkForCollision(UT_sint32 iIncomingPos, UT_sint32 iIncomingLength, UT_sint32 iIncomingAdjust, const UT_UTF8String& UUID, UT_sint32 iIncomingRemoteRev, UT_sint32& iRev, UT_sint32& iImportAdjustment)
{
	UT_DEBUGMSG(("ABI_Collab_Import::_checkForCollision() - iIncomingPos: %d, iIncomingLength: %d, UUID: %s, iIncomingRemoteRev: %d\n", iIncomingPos, iIncomingLength, UUID.utf8_str(), iIncomingRemoteRev));

	ABI_Collab_Export* pExport = m_pAbiCollab->getExport();
	UT_return_val_if_fail(pExport, false);

	const UT_GenericVector<ChangeAdjust *>* pExpAdjusts = pExport->getAdjusts();
	UT_return_val_if_fail(pExpAdjusts, false);

	iImportAdjustment = 0;

	// get the collision sequence (if any)
	UT_sint32 iStart = 0;
	UT_sint32 iEnd = 0;
	_calculateCollisionSeqence(iIncomingRemoteRev, UUID, iStart, iEnd);
	UT_return_val_if_fail(iStart >= 0 && iEnd >= 0, false);
	if (iStart == iEnd)
	{
		UT_DEBUGMSG(("Empty collision sequence, no possible collision\n"));
		return false;
	}

	std::deque<int> incAdjs;
	UT_sint32 iIncomingStateAdjust = _getIncomingAdjustmentForState(pExpAdjusts, iStart, iEnd, iIncomingPos, iIncomingLength, UUID, incAdjs);
	UT_DEBUGMSG(("IINCOMMINGSTATEADJUST: %d\n", iIncomingStateAdjust));
	UT_sint32 iIncomingStateAdjustUp = 0;

	// Now scan forward and look for an overlap of the new changerecord with the collision sequence
	UT_DEBUGMSG(("Checking collision sequence [%d..%d) for overlapping changerecords\n", iStart, iEnd));
	bool bDenied = false;
	for (UT_sint32 i = iStart; i < iEnd; i++)
	{
		ChangeAdjust* pChange = pExpAdjusts->getNthItem(i);
		if (pChange)
		{
			UT_DEBUGMSG(("Looking at pChange->m_pDocUUID: %s\n", pChange->m_pDocUUID));

			if (strcmp(pChange->m_pDocUUID, UUID.utf8_str()) != 0)
			{
				if (_isOverlapping(iIncomingPos+iIncomingStateAdjust+iIncomingStateAdjustUp, iIncomingLength, pChange->m_iDocPos, pChange->m_iLength))
				{
					bDenied = true;

					//
					// some overlaps are just fine; let's special case some here
					//

					if (iIncomingAdjust > 0 && pChange->m_iAdjust > 0)
					{
						// overlapping inserts are just fine, as long as the start positions differ
						if (iIncomingPos+iIncomingStateAdjust+iIncomingStateAdjustUp != static_cast<UT_sint32>(pChange->m_iDocPos))
						{
							UT_DEBUGMSG(("ALLOWING OVERLAPPING CHANGERECORD AFTER ALL\n"));
							bDenied = false;
						}
					}

					// TODO: add more overlaps, for example ones that change only (non-conflicting) formatting marks
				}

				if (bDenied)
				{
					UT_DEBUGMSG(("Overlap detected for iIncomingPos: %d, iIncomingLength: %d, pChange->m_iDocPos: %d, pChange->m_iLength: %d\n", 
							iIncomingPos, iIncomingLength, pChange->m_iDocPos, pChange->m_iLength));
					iRev = pChange->m_iCRNumber;
					break;
				}
				else
					UT_DEBUGMSG(("No overlap detected for iIncomingPos: %d, iIncomingLength: %d, pChange->m_iDocPos: %d, pChange->m_iLength: %d\n", 
							iIncomingPos, iIncomingLength, pChange->m_iDocPos, pChange->m_iLength));

				
				if (static_cast<UT_sint32>(pChange->m_iDocPos) < iIncomingPos+iIncomingStateAdjust+iIncomingStateAdjustUp)
				{
					UT_DEBUGMSG(("Normal Upward influence detected\n"));
					iIncomingStateAdjustUp += pChange->m_iAdjust;
				}
			}
			else
			{
				UT_DEBUGMSG(("Skipping overlap detection: changerecords came from the same document; iIncomingPos: %d, iIncomingAdjust: %d, iIncomingLength: %d, pChange->m_iDocPos: %d, pChange->m_iLength: %d\n", 
							iIncomingPos, iIncomingAdjust, iIncomingLength, pChange->m_iDocPos, pChange->m_iLength));
				iIncomingStateAdjustUp += incAdjs.front();
				incAdjs.pop_front();
			}

			UT_DEBUGMSG(("Now: iIncomingStateAdjust: %d, iIncomingStateAdjustUp: %d - (total import adjustment: %d) \n", iIncomingStateAdjust, iIncomingStateAdjustUp, iIncomingStateAdjust+iIncomingStateAdjustUp));
		}
		else
			UT_return_val_if_fail(false, false);
	}

	while (!incAdjs.empty())
	{
		UT_DEBUGMSG(("Adding left-over incoming adjustment: %d\n", incAdjs.front()));
		iIncomingStateAdjustUp += incAdjs.front();
		incAdjs.pop_front();
	}

	iImportAdjustment = iIncomingStateAdjust+iIncomingStateAdjustUp;
	UT_DEBUGMSG(("Full import adjustment: %d\n", iImportAdjustment));

	return bDenied;
}

// returns true if the import can continue, false otherwise
bool ABI_Collab_Import::_handleCollision(UT_sint32 iIncomingRev, UT_sint32 iLocalRev, const Buddy& collaborator)
{
	UT_DEBUGMSG(("_handleCollision() - incoming rev %d collides against local rev %d!!!\n", iIncomingRev, iLocalRev));

	if (m_pAbiCollab->isLocallyControlled())
	{
		UT_DEBUGMSG(("We're controlling this session, refusing this changerecord from %s!\n", collaborator.getName().utf8_str()));
		// add this collaborator to our revert ack list, so we can ignore his packets
		// until we get an acknoledgement that he has reverted his local, colliding changes
		m_revertSet.push_back(collaborator.getName());
		// send the revert command to the collaborator
		RevertSessionPacket rsp(m_pAbiCollab->getSessionId(), m_pDoc->getOrigDocUUIDString(), iIncomingRev);
		m_pAbiCollab->push(&rsp, collaborator);
		return false;
	}
	else
	{
		UT_DEBUGMSG(("We're NOT controlling this session, reverting local changes and accepting changerecord!\n"));

		ABI_Collab_Export* pExport = m_pAbiCollab->getExport();
		UT_return_val_if_fail(pExport, false);

		UT_GenericVector<ChangeAdjust *>* pExpAdjusts = pExport->getAdjusts();
		UT_return_val_if_fail(pExpAdjusts, false);
		
		m_pAbiCollab->setIsReverting(true); // mask all changes in the exporter

		// undo our cool local changes, and nuke our exported packet list as well up to (and including) iLocalRev
		for (UT_sint32 i = pExpAdjusts->getItemCount() - 1; i >= 0; i--)
		{
			ChangeAdjust* pChange = pExpAdjusts->getNthItem(i);
			if (pChange)
			{
				if (pChange->m_iCRNumber >= iLocalRev)
				{
					if (strcmp(m_pDoc->getOrigDocUUIDString(), pChange->m_pDocUUID) == 0)
					{
						UT_DEBUGMSG(("UNDO-ING AND NUKING LOCAL CHANGE: EXPORT POSITION %d, pChange->m_iCRNumber: %d!\n", i, pChange->m_iCRNumber));

						// undo the change locally
						m_pDoc->undoCmd(1);

						// fix up the positions on the change stack
						for (UT_uint32 j = i+1; j < pExpAdjusts->getItemCount(); j++)
						{
							ChangeAdjust* pC = pExpAdjusts->getNthItem(j);
							if (pC)
							{
								UT_DEBUGMSG(("Looking at fixing up the position of change pos %d\n", j));
								if (pChange->m_iDocPos < pC->m_iDocPos)
								{
									UT_DEBUGMSG(("Adjusting change pos %d from m_iDocPos: %d to m_iDocPos: %d\n", j, pC->m_iDocPos, pC->m_iDocPos - pChange->m_iAdjust));
									pC->m_iDocPos = pC->m_iDocPos - pChange->m_iAdjust;
								}
								else
									UT_DEBUGMSG(("No need to adjust change pos %d\n", j));
							}
							else
								UT_ASSERT_HARMLESS(UT_SHOULD_NOT_HAPPEN);
						}
						
						// kill off the item
						pExpAdjusts->deleteNthItem(i);
						delete pChange;
					}
					else
						UT_DEBUGMSG(("Skipping undo of remote change\n"));
				}
				else
					break;
			}
			else
				UT_ASSERT_HARMLESS(UT_SHOULD_NOT_HAPPEN);
		}

		m_pAbiCollab->setIsReverting(false); // unmask all changes in the exporter

		UT_DEBUGMSG(("Pre-Acknowledging revert of revision %d\n", iLocalRev));
		// send the revert acknowledgement command to the session owner
		RevertAckSessionPacket rasp(m_pAbiCollab->getSessionId(), m_pDoc->getOrigDocUUIDString(), iLocalRev);
		m_pAbiCollab->push( &rasp, collaborator );

		m_iAlreadyRevertedRevs.push_back(iLocalRev);	
		
		return true;
	}
}

bool ABI_Collab_Import::_shouldIgnore(const Buddy& collaborator)
{
	if (m_pAbiCollab->isLocallyControlled())
	{
		UT_DEBUGMSG(("This session is locally controlled, check if we are waiting for a revert ack from buddy: %s\n", collaborator.getName().utf8_str()));
		// see if we are waiting for a revert ack packet from this collaborator;
		// if we do, then just drop all packets on the floor until we see it
		for (vector<UT_UTF8String>::iterator it = m_revertSet.begin(); it != m_revertSet.end(); it++)
		{
			if (*it == collaborator.getName())
			{
				UT_DEBUGMSG(("Found collaborator %s on our revert ack list; changerecords should be ignored!\n", collaborator.getName().utf8_str()));
				return true;
			}
		}
	}
	UT_DEBUGMSG(("%s is not on our revert ack list, don't ignore this packet...\n", collaborator.getName().utf8_str()));
	return false;
}

bool ABI_Collab_Import::import(const SessionPacket& packet, const Buddy& collaborator)
{
	UT_DEBUGMSG(("ABI_Collab_Import::import()\n"));

	UT_DEBUGMSG(("--------------------------\n"));
	UT_DEBUGMSG(("%s", packet.toStr().c_str()));
	UT_DEBUGMSG(("--------------------------\n"));
	
	// check for collisions to see if we can import this packet at all;
	// NOTE: the position adjustment is calculated in the process, so we don't have to 
	// do that again on import (as calculating it for globs can be quite costly, and
	// we need to do it anyway to properly detect collisions)
	UT_sint32 iImportAdjustment = 0;
	switch (packet.getSessionType())
	{
		case PACKET_GLOB:
			{
				if (_shouldIgnore(collaborator))
					return false;

				const GlobSessionPacket* gp = static_cast<const GlobSessionPacket*>(&packet);
				UT_return_val_if_fail(gp->getPackets().size() > 0, false);

				UT_DEBUGMSG(("Determining collision for glob as a whole: iGlobPos: %d, iGlobLength: %d, iGlobRemoteRev: %d, iGlobRev: %d\n", gp->getPos(), gp->getLength(), gp->getRemoteRev(), gp->getRev()));

				UT_sint32 iLocalRev = 0;
				bool bCollide = _checkForCollision(gp->getPos(), gp->getLength(), gp->getAdjust(), gp->getDocUUID(), gp->getRemoteRev(), iLocalRev, iImportAdjustment);
				// make sure we revert all the way upto the first revision in this glob if a collision is detected
				bool continueImport = (bCollide	? _handleCollision(gp->getRev(), iLocalRev, collaborator) : true);
				if (!continueImport)
					return false;	

				m_pAbiCollab->setOrigDocPos(gp->getPos()); // HACK HACK HACK
				m_pAbiCollab->setOrigRemoteSeen(gp->getRemoteRev()); // HACK HACK HACK
			}
			break;
		case PACKET_CHANGERECORD:
			{
				if (_shouldIgnore(collaborator))
					return false;

				// check for a collision and handle it if it occurred
				const ChangeRecordSessionPacket* crp = static_cast<const ChangeRecordSessionPacket*>(&packet);
				UT_DEBUGMSG(("%s\n", crp->toStr().c_str()));
				UT_sint32 iLocalRev = 0;
				bool bCollide = _checkForCollision(crp->getPos(), crp->getLength(), crp->getAdjust(), crp->getDocUUID(), crp->getRemoteRev(), iLocalRev, iImportAdjustment);
				bool continueImport = (bCollide ? _handleCollision(crp->getRev(), iLocalRev, collaborator) : true);
				if (!continueImport)
					return false;

				m_pAbiCollab->setOrigDocPos(crp->getPos()); // HACK HACK HACK
				m_pAbiCollab->setOrigRemoteSeen(crp->getRemoteRev()); // HACK HACK HACK
			}
			break;

		case PACKET_SIGNAL:
			if (_shouldIgnore(collaborator))
				return false;
			// this packet can never cause a collision, let it pass
			break;
		case PACKET_REVERT:
		case PACKET_REVERT_ACK:
			// these packets can never cause collisions, let them pass
			break;
		default:
			UT_DEBUGMSG(("Unimplemented packet session type: %d", packet.getSessionType()));
			UT_ASSERT_HARMLESS(UT_NOT_IMPLEMENTED);
			break;
	}

	UT_DEBUGMSG(("Collision test succeeded, continuing to import this packet\n"));

	UT_GenericVector<AV_View *> vecViews;
	m_pDoc->getAllViews(&vecViews);

	for (UT_uint32 i=0; i < vecViews.getItemCount(); i++)
	{
		vecViews.getNthItem(i)->setActivityMask(false);
	}

	// import the changes in the document
	UT_UTF8String sRealDocname = m_pDoc->getOrigDocUUIDString();
	bool bRes = _import(packet, iImportAdjustment, collaborator);
	m_pDoc->setMyUUID(sRealDocname.utf8_str());
	UT_DEBUGMSG(("Resetting orig docpos && orig remote seen\n"));
	m_pAbiCollab->setOrigDocPos(-1); // HACK HACK HACK
	m_pAbiCollab->setOrigRemoteSeen(-1); // HACK HACK HACK

	bRes = true; // force nice updates for now; stuff like tables still report false on import, due to formatting marks without props/attrs
	if (bRes)
	{
		bool bDone = false;
		for (UT_uint32 i = 0; i<vecViews.getItemCount(); i++)
		{
			FV_View * pView = static_cast<FV_View *>( vecViews.getNthItem(i));
			if(pView && !bDone && pView->shouldScreenUpdateOnGeneralUpdate())
			{
				m_pDoc->signalListeners(PD_SIGNAL_UPDATE_LAYOUT);
				bDone = true;
			}
			if(pView)
			{
				pView->fixInsertionPointCoords();
				pView->setActivityMask(true);
			}
		}
	}
	else
	{
		for (UT_uint32 i = 0; i<vecViews.getItemCount(); i++)
		{
			vecViews.getNthItem(i)->setActivityMask(true);
		}
	}

	return bRes;
}

/*!
 * Take a packet contained with a UT_UTF8string, interpret it's
 * contents and apply the implied operations on the document.
 */
bool ABI_Collab_Import::_import(const SessionPacket& packet, UT_sint32 iImportAdjustment, const Buddy& collaborator, bool inGlob)
{
	UT_DEBUGMSG(("ABI_Collab_Import::_import() - packet type: %d, iImportAdjustment: %d\n", packet.getSessionType(), iImportAdjustment));

	switch (packet.getSessionType())
	{
		case PACKET_GLOB:
			{
				const GlobSessionPacket* gp = static_cast<const GlobSessionPacket*>(&packet);
				UT_return_val_if_fail(gp->getPackets().size() > 0, false);

				// store the last seen revision from this collaborator (it is immediately used by the export)
				m_remoteRevs[collaborator.getName().utf8_str()] = gp->getRev();

				// lock out all updates while processing this glob
				m_pDoc->notifyPieceTableChangeStart();
				m_pDoc->disableListUpdates();
				m_pDoc->setDontImmediatelyLayout(true);
				m_pDoc->beginUserAtomicGlob();

				for (UT_uint32 j = 0; j < gp->getPackets().size(); j++)
				{
					SessionPacket* pGlobPacket = gp->getPackets()[j];
					if (pGlobPacket)
					{
						bool res = _import(*pGlobPacket, iImportAdjustment, collaborator, true); // yay for recursion :)
						if (!res)
							UT_ASSERT_HARMLESS(UT_SHOULD_NOT_HAPPEN);
					}
				}

				// allow updates again
				m_pDoc->enableListUpdates();
				m_pDoc->updateDirtyLists();
				m_pDoc->notifyPieceTableChangeEnd();
				m_pDoc->setDontImmediatelyLayout(false);
				m_pDoc->endUserAtomicGlob();

				return true;
			}
		
		case PACKET_SIGNAL:
			{
				const SignalSessionPacket* sp = static_cast<const SignalSessionPacket*>(&packet);
				// TODO: FIX ME FIX ME FIXME
				if(0 /*m_pAbiCollab->isServer() */ )
				{
					if( sp->getSignalType() == PD_SIGNAL_DOCSAVED)
					{
						m_pDoc->save();
						return true;
					}
				}
				if (sp->getSignalType() == PD_SIGNAL_DOCCLOSED)
				{
					m_pAbiCollab->setClose();
					UT_DEBUGMSG(("Doing close from remote signal \n"));
					return true;
				}
				m_pDoc->signalListeners(sp->getSignalType());
				return true;
			}

		case PACKET_REVERT:
			{
				const RevertSessionPacket* rrp = static_cast<const RevertSessionPacket*>(&packet);
				UT_DEBUGMSG(("Revert packet seen on import for rev: %d\n", rrp->getRev()));

				if (m_iAlreadyRevertedRevs.size() == 0 || m_iAlreadyRevertedRevs.front() != rrp->getRev())
				{
					UT_DEBUGMSG(("Incoming revert for revision %d, which we didn't detect locally (m_iAlreadyRevertedRev: %d)!\n", rrp->getRev(), m_iAlreadyRevertedRevs.front()));
					UT_DEBUGMSG(("DOCUMENT OUT OF SYNC DETECTED!!!!\n"));
					UT_ASSERT_HARMLESS(UT_NOT_IMPLEMENTED);
					return false;
				}	

				m_iAlreadyRevertedRevs.pop_front();
				return true;
			}
		case PACKET_REVERT_ACK:
			{
				UT_DEBUGMSG(("RevertAck packet seen on import for rev: %d\n", static_cast<const RevertAckSessionPacket*>(&packet)->getRev()));

				// TODO: check the rev to make sure it matches the one we sent out

				// remove this collaborator from our revert ack list; he can play again...
				for (vector<UT_UTF8String>::iterator it = m_revertSet.begin(); it != m_revertSet.end(); it++)
				{
					if (*it == collaborator.getName())
					{
						UT_DEBUGMSG(("Found collaborator %s on our revert ack list! Removing him from the list...\n", (*it).utf8_str()));
						m_revertSet.erase(it);
						return true;
					}
				}

				UT_ASSERT_HARMLESS(UT_SHOULD_NOT_HAPPEN);
				return false;
			}

		case PACKET_CHANGERECORD:
			{
				const ChangeRecordSessionPacket* crp = static_cast<const ChangeRecordSessionPacket*>(&packet);
				UT_DEBUGMSG(("It's safe to import this packet\n"));

				UT_DEBUGMSG(("For CR number %d requested point %d adjustment %d \n",  crp->getRev(), crp->getPos(), iImportAdjustment));

				PT_DocPosition pos = static_cast<PT_DocPosition>(crp->getPos() + iImportAdjustment);
				UT_ASSERT(pos <= getEndOfDoc());

				if (!inGlob)
				{
					// store the last seen revision from this collaborator (it is immediately used by the export)
					// NOTE: if this changerecord is part of a glob, then we don't do this; we'll have
					// already set the revision of the glob itself as the last seen one
					m_remoteRevs[collaborator.getName().utf8_str()] = crp->getRev();
				}

				// todo: remove these temp vars
				PT_DocPosition iPos2 = 0;
				m_pDoc->setMyUUID(crp->getDocUUID().utf8_str());

				// process the packet
				switch(crp->getPXType())
				{
					case PX_ChangeRecord::PXT_GlobMarker:
					{
						UT_DEBUGMSG(("Found GLOB marker (ignoring)\n"));
						return true;
					}
					case PX_ChangeRecord::PXT_InsertSpan:
					{
						const InsertSpan_ChangeRecordSessionPacket* icrsp = static_cast<const InsertSpan_ChangeRecordSessionPacket*>( crp );
						UT_UCS4String UCSChars = const_cast<UT_UTF8String&>(icrsp->m_sText).ucs4_str();	// ugly, ucs4_str should be const func!
						PP_AttrProp attrProp;
						attrProp.setAttributes(const_cast<const gchar**>(icrsp->getAtts()));
						attrProp.setProperties(const_cast<const gchar**>(icrsp->getProps()));
						m_pDoc->insertSpan(pos,UCSChars.ucs4_str(),UCSChars.length(), &attrProp);
						break;
					}
					case PX_ChangeRecord::PXT_DeleteSpan:
					{
						iPos2 = pos + crp->getLength();
						PP_AttrProp *p_AttrProp_Before = NULL;
						UT_uint32 icnt = 0;
						m_pDoc->deleteSpan(pos,iPos2,p_AttrProp_Before,icnt,true);
						break;
					}
					case PX_ChangeRecord::PXT_ChangeSpan:
					{
						const Props_ChangeRecordSessionPacket* pcrsp = static_cast<const Props_ChangeRecordSessionPacket*>( crp );
						gchar** szAtts = pcrsp->getAtts();
						gchar** szProps = pcrsp->getProps();
						iPos2 = pos + pcrsp->getLength();
						if((szProps == NULL) && (szAtts == NULL))
						{
							//
							// This happens if we remove all formats
							// we have to handle this seperately
							//
							// Get style of containing block
							//
							PL_StruxDocHandle sdh = NULL;
							m_pDoc->getStruxOfTypeFromPosition(pos,PTX_Block,&sdh);
							if(!sdh)
							{
								UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
								return false;
							}
							PD_Style * pStyle = m_pDoc->getStyleFromSDH(sdh);
							if(!pStyle)
							{
								UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
								return false;
							}
							const char * szName =  pStyle->getName();
							const char * atts[3] = {PT_STYLE_ATTRIBUTE_NAME,szName,NULL};
							m_pDoc->changeSpanFmt(PTC_SetExactly, pos, iPos2, atts, const_cast<const gchar**>( szProps ) );
						}
						else
						{
							m_pDoc->changeSpanFmt(PTC_SetExactly, pos, iPos2, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ) );
						}
						break;
					}
					case PX_ChangeRecord::PXT_InsertStrux:
					{
						const ChangeStrux_ChangeRecordSessionPacket* pcrsp = static_cast<const ChangeStrux_ChangeRecordSessionPacket*>( crp );
						PTStruxType pts = pcrsp->m_eStruxType;
						gchar** szAtts = pcrsp->getAtts();
						gchar** szProps = pcrsp->getProps();
						
						if((szProps != NULL) || (szAtts != NULL))
						{
							m_pDoc->insertStrux( pos, pts, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ) );
						}
						else
						{
							m_pDoc->insertStrux(pos, pts);
						}
						break;
					}
					case PX_ChangeRecord::PXT_DeleteStrux:
					{
						const DeleteStrux_ChangeRecordSessionPacket* pcrsp = static_cast<const DeleteStrux_ChangeRecordSessionPacket*>( crp );
						PTStruxType pts = pcrsp->m_eStruxType;
						m_pDoc->deleteStrux(pos,pts,true);
						break;
					}
					case PX_ChangeRecord::PXT_ChangeStrux:
					{
						const ChangeStrux_ChangeRecordSessionPacket* pcrsp = static_cast<const ChangeStrux_ChangeRecordSessionPacket*>( crp );
						PTStruxType pts = pcrsp->m_eStruxType;
						gchar** szAtts = pcrsp->getAtts();
						gchar** szProps = pcrsp->getProps();
						UT_return_val_if_fail(szProps != NULL || szAtts != NULL, false);

						UT_DEBUGMSG(("Executing ChangeStrux pos= %d \n",pos));
						m_pDoc->changeStruxFmt(PTC_SetExactly, pos, pos, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ), pts);
						
						// TODO: this mask is waaaay to generic
						XAP_Frame *pFrame = XAP_App::getApp()->getLastFocussedFrame();
						FV_View* pView = static_cast<FV_View*>(pFrame->getCurrentView());
						pView->notifyListeners(AV_CHG_TYPING | AV_CHG_FMTCHAR | AV_CHG_FMTBLOCK | AV_CHG_PAGECOUNT | AV_CHG_FMTSTYLE );
						break;
					}
					case PX_ChangeRecord::PXT_InsertObject:
					{
						const Object_ChangeRecordSessionPacket* ocrsp = static_cast<const Object_ChangeRecordSessionPacket*>( crp );
						PTObjectType pto = ocrsp->m_eObjectType;
						gchar** szAtts = ocrsp->getAtts();
						gchar** szProps = ocrsp->getProps();
						
						if((szProps == NULL) && (szAtts == NULL))
						{
							UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
							return false;
						}
						m_pDoc->insertObject(pos, pto, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ) );
						break;
					}
					case PX_ChangeRecord::PXT_DeleteObject:
					{
						iPos2 = pos + 1;
						PP_AttrProp *p_AttrProp_Before = NULL;
						UT_uint32 icnt = 0;
						m_pDoc->deleteSpan(pos, iPos2, p_AttrProp_Before, icnt, true);
						break;
					}
					case PX_ChangeRecord::PXT_ChangeObject:
					{
						const Object_ChangeRecordSessionPacket* ccrsp = static_cast<const Object_ChangeRecordSessionPacket*>( crp );
						//PTObjectType pto = ccrsp->m_eObjectType;
						gchar** szAtts = ccrsp->getAtts();
						gchar** szProps = ccrsp->getProps();
						
						if ((szProps == NULL) && (szAtts == NULL))
						{
							UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
							return false;
						}
						m_pDoc->changeSpanFmt(PTC_SetExactly, pos, pos + 1, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ));
						break;
					}
					case PX_ChangeRecord::PXT_InsertFmtMark:
					{
						const Props_ChangeRecordSessionPacket* pcrsp = static_cast<const Props_ChangeRecordSessionPacket*>( crp );
						gchar** szAtts = pcrsp->getAtts();
						gchar** szProps = pcrsp->getProps();
						
						if((szProps == NULL) && (szAtts == NULL))
						{
							// nothing to do here, please move along
							
							// NOTE: why does this happen anyway? 
							// This happens when for example when sending over tables:
							UT_ASSERT_HARMLESS(UT_SHOULD_NOT_HAPPEN);
							return false;
						}
						return m_pDoc->changeSpanFmt(PTC_SetExactly, pos, pos, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ));
					}
					case PX_ChangeRecord::PXT_DeleteFmtMark:
					{
						return m_pDoc->deleteFmtMark(pos);
					}
					case PX_ChangeRecord::PXT_ChangeFmtMark:
					{
						const Props_ChangeRecordSessionPacket* pcrsp = static_cast<const Props_ChangeRecordSessionPacket*>( crp );
						gchar** szAtts = pcrsp->getAtts();
						gchar** szProps = pcrsp->getProps();
						
						if ((szProps == NULL) && (szAtts == NULL))
						{
							UT_ASSERT(UT_SHOULD_NOT_HAPPEN);
							return false;
						}
						return m_pDoc->changeSpanFmt(PTC_SetExactly, pos, pos, const_cast<const gchar**>( szAtts ), const_cast<const gchar**>( szProps ));
					}
					case PX_ChangeRecord::PXT_ChangePoint:
					{
						UT_DEBUGMSG(("Change Point CR \n"));
						return m_pDoc->createAndSendCR(pos, crp->getPXType(), true, 0);
					}
					case PX_ChangeRecord::PXT_ListUpdate:
					{
						UT_DEBUGMSG(("ListUpdate CR \n"));
						return m_pDoc->createAndSendCR(pos, crp->getPXType(), true,0);
					}
					case PX_ChangeRecord::PXT_StopList:
					{
						UT_DEBUGMSG(("StopList CR \n"));
						return m_pDoc->createAndSendCR(pos, crp->getPXType(), true,0);
					}
					case PX_ChangeRecord::PXT_UpdateField:
					{
						UT_DEBUGMSG(("UpdateFiled CR \n"));
						return m_pDoc->createAndSendCR(pos, crp->getPXType(), true,0);
					}
					case PX_ChangeRecord::PXT_RemoveList:
					{
						UT_DEBUGMSG(("RemoveList CR \n"));
						return m_pDoc->createAndSendCR(pos, crp->getPXType(), true,0);
					}
					case PX_ChangeRecord::PXT_UpdateLayout:
					{
						UT_DEBUGMSG(("UpdateLayout CR \n"));
						return m_pDoc->createAndSendCR(pos, crp->getPXType(), true,0);
					}
					case PX_ChangeRecord::PXT_CreateDataItem:
					{
						const Data_ChangeRecordSessionPacket* dp = static_cast<const Data_ChangeRecordSessionPacket*>( crp );
						const char * szNameV = g_strdup(dp->getAttribute(PT_DATAITEM_ATTRIBUTE_NAME));
						void * pHandle = NULL;
						char * pToken = dp->m_bTokenSet ? g_strdup(dp->m_sToken.c_str()) : NULL;
						UT_ByteBuf * pBuf= new UT_ByteBuf();
						UT_DEBUGMSG(("PXT_CreateDataItem: append image buffer @ 0x%x, %u bytes, pToken %s\n", &dp->m_vecData[0], dp->m_vecData.size(), pToken?pToken:"NULL"));
						pBuf->append(reinterpret_cast<const UT_Byte *>( &dp->m_vecData[0] ), dp->m_vecData.size() );
						bool res = m_pDoc->createDataItem(szNameV,false,pBuf,pToken,&pHandle);
						delete pBuf;
						return res;
					}
					default:
					{
						UT_DEBUGMSG(("Unimplemented crp->getPXType(): %d\n", crp->getPXType()));
						UT_ASSERT_HARMLESS(UT_NOT_IMPLEMENTED);
						break;
					}
				}
			}
			return true;			
		
		default:
			UT_DEBUGMSG(("ABI_Collab_Import::import called with wrong packet type: %d!\n", packet.getType()));
			return false;			
	}

	return false;
}

bool ABI_Collab_Import::getStruxType(UT_UTF8String & sInPacket, 
				     PTStruxType & struxType,
				     UT_UTF8String & sAttributes, 
				     UT_UTF8String & sOutpacket)
{
	return true;
}

bool ABI_Collab_Import::getObjectType(UT_UTF8String & sInPacket,
				      PTObjectType & objectType,
				      UT_UTF8String & sAttributes, 
				      UT_UTF8String & sOutpacket)
{
	return true;
}

PT_DocPosition ABI_Collab_Import::getEndOfDoc(void)
{
        PT_DocPosition posEnd;
        m_pDoc->getBounds(true,posEnd);
        return posEnd;
}
