/*
 * AbiCollab - Code to enable the modification of remote documents.
 * Copyright (C) 2005 by Martin Sevior
 * Copyright (C) 2006 by Marc Maurer <uwog@uwog.net>
 * Copyright (C) 2007 One Laptop Per Child
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef ABI_DISK_RECORDER_H
#define ABI_DISK_RECORDER_H

#include <string>
#include "ut_go_file.h"
#include "AbiCollab_Plugin.h"
#include <time.h>

class ABI_EXPORT DiskSessionRecorder : public SessionRecorderInterface
{
public:
	static const char* getTargetDirectory()
	{
		return XAP_App::getApp()->getUserPrivateDirectory();
	}
	static const char* getPrefix()
	{
		return "Session-";
	}
		
public:
	DiskSessionRecorder( AbiCollab* Session ) 
	: SessionRecorderInterface( Session )
	{
		UT_DEBUGMSG(("DiskSessionRecorder::DiskSessionRecorder\n"));
		
		std::string pidStr;
#ifndef WIN32
		pidStr = str(boost::format( "%1%" ) % int(getpid()) );
#endif
		gchar *s = g_build_filename( getTargetDirectory(), (std::string(getPrefix())+Session->getSessionId().utf8_str()).c_str(), NULL );
		std::string fn = (char*)s;
		fn += ".";
		fn += pidStr;
		FREEP(s);
		
		m_URI = UT_go_filename_to_uri(fn.c_str());
		m_Error = 0;
		m_File = UT_go_file_create( m_URI, &m_Error );

		if (m_File) {
			UT_DEBUGMSG(("DiskSessionRecorder: writing to file %s\n", fn.c_str()));
			write( getHeader(), strlen(getHeader()) );
			write( ABICOLLAB_PROTOCOL_VERSION, 1 );
		} else {
			UT_DEBUGMSG(("DiskSessionRecorder: stopping, can't write to file %s\n", fn.c_str()));
		}
	}
	~DiskSessionRecorder() 
	{
		destroy();
	}
	void storeOutgoing( const Packet* pPacket )
	{
		store( false, pPacket, NULL );
	}
	void storeOutgoing( const Packet* pPacket, const Buddy& ToBuddy )
	{
		store( false, pPacket, &ToBuddy );
	}
	void storeIncoming( const Packet* pPacket, const Buddy& FromBuddy )
	{
		store( true, pPacket, &FromBuddy );
	}
	static const char* getHeader()
	{
		return "DSR!";
	}
	
protected:
	GsfOutput* 		m_File;
	GError*			m_Error;
	const char*		m_URI;
	
	void destroy()
	{
		UT_DEBUGMSG(("DiskSessionRecorder::destroy: closing file 0x%x\n", m_File));
		if (m_File) 
		{
			gsf_output_close(m_File);
			g_object_unref(G_OBJECT(m_File));
			m_File = NULL;	// necessary? /me is gsf n00b
		}
		if (m_URI) 
		{
			FREEP(m_URI);
		}
	}
	void store( bool incoming, const Packet* pPacket, const Buddy* pBuddy )
	{
		UT_DEBUGMSG(("DiskSessionRecorder::store: %d 0x%x %s\n", incoming, pPacket, pBuddy?pBuddy->getName().utf8_str():"none"));
		UT_return_if_fail(pPacket);
		UT_return_if_fail(m_File);
		
		OStrArchive os;
		
		// store if we're incoming or outgoing
		char incomingC = incoming;
		os << incomingC;

		// store if we had a buddy, and who it was
		char haveBuddy = pBuddy ? 1 : 0;
		os << haveBuddy;
		if (haveBuddy) {
			os << const_cast<UT_UTF8String&>( pBuddy->getName() );
		}
		
		// store timestamp, make it 64-bit value always
		UT_uint64 timestamp = UT_uint64( time(0) );
		os << timestamp;

		// store packet class
		unsigned char packetClass = pPacket->getClassType();
		os << packetClass;

		// store packet
		os << const_cast<Packet&>( *pPacket );

		// write to file!
		write( os.getData().c_str(), os.Size() );
	}
	void write( const void* data, int count )
	{
		UT_DEBUGMSG(("DiskSessionRecorder::write: 0x%x %d\n", data,count));
		UT_return_if_fail(m_File);
		gsf_output_write( m_File, size_t(count), reinterpret_cast<const guint8*>( data ) );
	}
};

#endif
