/*
 * AbiCollab - Code to enable the modification of remote documents.
 * Copyright (C) 2005 by Martin Sevior
 * Copyright (C) 2006 by Marc Maurer <uwog@uwog.net>
 * Copyright (C) 2007 One Laptop Per Child
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

// not exporting anything here
#define ABI_EXPORT

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <time.h>

#include <ut_string_class.h>
#include "xap_App.h"
#include <backends/xp/Buddy.h>
#include <xp/AbiCollab.h>
#include <xp/AbiCollab_Plugin.h>

#include <xp/Serialization.h>
#include "DiskSessionRecorder.h"

bool DumpSession( const std::string& Filename )
{
	// open file
	GsfInput* in = UT_go_file_open(Filename.c_str(), NULL);
	UT_return_val_if_fail(in,false);
	
	// read contents
	size_t fileSize = gsf_input_size(in);
	guint8 const* contents = gsf_input_read(in, fileSize, NULL);
	if (!contents) {
		g_object_unref(G_OBJECT(in));
		UT_return_val_if_fail(false,false);
	}
	
	// ugly: copy to string, because our serialization code expects a string
	std::string buffer;
	buffer.resize( fileSize );
	memcpy( &buffer[0], contents, fileSize );
	
	// check header
	size_t headerLen = strlen( DiskSessionRecorder::getHeader() );
	if (memcmp( DiskSessionRecorder::getHeader(), &buffer[0], headerLen )) 
	{
		UT_DEBUGMSG(("%s does not seem to be a session file\n", Filename.c_str() ));
		UT_return_val_if_fail(false,false);
	}
	
	// check version
	size_t versionLen = strlen( ABICOLLAB_PROTOCOL_VERSION );
	if (memcmp( ABICOLLAB_PROTOCOL_VERSION, &buffer[headerLen], versionLen ))
	{
		UT_DEBUGMSG(("%s has wrong version\n", Filename.c_str() ));
		UT_return_val_if_fail(false,false);
	}
	
	// create stream, and skip header (is non-serialized-data)
	IStrArchive is( buffer );
	is.Skip( headerLen + versionLen );
	
	// keep reading packets
	try
	{
		size_t packetCounter = 0;
		while (!is.EndOfFile()) 
		{
			// read data for packet
			char incoming;
			is << incoming;
			
			char haveBuddy;
			is << haveBuddy;
			
			UT_UTF8String buddyName;
			if (haveBuddy) {
				is << buddyName;
			}
			
			UT_uint64 timestamp;
			is << timestamp;
			time_t t = time_t(timestamp);
			struct tm time;
			gmtime_r( &t, &time );
			
			unsigned char packetClass;
			is << packetClass;
			Packet* newPacket = Packet::createPacket( (PClassType)packetClass );
			if (!newPacket) {
				UT_DEBUGMSG(("unknown packet class %d\n", packetClass ));
				UT_return_val_if_fail(false,false);
			}
			is << *newPacket;
			
			// display packet
			printf("--------------------------------------------------------------------------------\n");
			printf("@ %04d/%02d/%02d %02d:%02d:%02d\n", 1900+time.tm_year, time.tm_mon, time.tm_mday, time.tm_hour, time.tm_min, time.tm_sec);
			printf("[%06u] %s packet ", packetCounter++, incoming?"INCOMING":"OUTGOING" );
			printf("%s ", incoming?"from":"to");
			if (haveBuddy) {
				printf("<%s>", buddyName.utf8_str());
			} else {
				printf("<all>");
			}
			printf(" of class %s\n", Packet::getPacketClassname((PClassType)packetClass));
			printf("--------------------------------------------------------------------------------\n");
			printf("%s\n", newPacket->toStr().c_str());
				
			// cleanup
			delete newPacket;
		}
		printf("--------------------------------------------------------------------------------\n");
	}
	catch(...)
	{
		UT_DEBUGMSG(("exception caught during packet serialization\n"));
	}
	return true;
}

/*
 * 4-left
 */
